﻿using System;
using System.Linq;
using System.Net;
using System.Xml.Linq;
using SitemapUpdateNotification.Configuration;

namespace SitemapUpdateNotification
{
    /// <summary>
    /// Contains methods for notifying configured targets, such as search engines,
    /// when a specified XML sitemap has been updated.
    /// </summary>
    public static class SitemapUpdateNotifier
    {
        private const string SITEMAP_NAMESPACE = "http://www.sitemaps.org/schemas/sitemap/0.9";

        /// <summary>
        /// Pings all configured notification targets with the configured sitemap.
        /// </summary>
        /// <returns>A summary of which notification targets that where successfully pinged
        /// and any that where not.</returns>
        public static string UpdatedSitemapNotify()
        {
            return UpdatedSitemapNotify(null);
        }

        /// <summary>
        /// Pings all configured notification targets with the configured sitemap
        /// if the sitemap has been updated after a specified date.
        /// </summary>
        /// <param name="updatedAfter">Notification targets will be notified 
        /// about the sitemap if it has been updated after this date.</param>
        /// <returns>A summary of which notification targets that where successfully pinged
        /// and any that where not.</returns>
        public static string UpdatedSitemapNotify(DateTime? updatedAfter)
        {
            string rootUrl = Configuration.SitemapUrl;

            if (!updatedAfter.HasValue)
                return Ping(rootUrl);

            XDocument rootDocument = XDocument.Load(rootUrl);

            string result;

            if (IsUpdated(rootDocument, updatedAfter))
            {
                result = Ping(rootUrl);
            }
            else
            {
                result = string.Format("No updates made to the sitemap after {0}.", updatedAfter.Value.ToString("yyyy-MM-dd hh:mm:ss"));
            }

            return result;
        }

        private static bool IsUpdated(XDocument rootDocument, DateTime? updatedAfter)
        {
            bool updated = false;

            if (IsIndexDocument(rootDocument))
            {
                if (PointsToUpdatedSitemap(rootDocument, updatedAfter.Value))
                    updated = true;
            }
            else
            {
                if (UpdatedAfter(rootDocument, updatedAfter.Value))
                    updated = true;
            }
            return updated;
        }

        private static bool IsIndexDocument(XDocument sitemapIndexDocument)
        {
            return sitemapIndexDocument.Descendants(SitemapNamespace + "sitemap").Count() > 0;
        }

        private static bool PointsToUpdatedSitemap(XDocument rootDocument, DateTime updatedAfter)
        {
            var sitemapUrls = from sitemap in rootDocument.Descendants(SitemapNamespace + "sitemap")
                              where sitemap.Element(SitemapNamespace + "loc") != null
                              select sitemap.Element(SitemapNamespace + "loc").Value;

            bool updated = false;

            foreach (string sitemapUrl in sitemapUrls)
            {
                XDocument sitemapDocument = XDocument.Load(sitemapUrl);

                if (UpdatedAfter(sitemapDocument, updatedAfter))
                {
                    updated = true;
                    break;
                }

            }

            return updated;
        }

        private static bool UpdatedAfter(XDocument sitemapDocument, DateTime updatedAfter)
        {
            var modificationDates = from url in sitemapDocument.Descendants(SitemapNamespace + "url")
                                    where url.Element(SitemapNamespace + "lastmod") != null
                                    orderby DateTime.Parse(url.Element(SitemapNamespace + "lastmod").Value) descending
                                    select DateTime.Parse(url.Element(SitemapNamespace + "lastmod").Value);

            if (modificationDates.Count() == 0)
                return false;

            DateTime lastModificationDate = modificationDates.First();

            if (lastModificationDate > updatedAfter)
                return true;

            return false;
        }

        private static string Ping(string sitemapUrl)
        {
            string result = string.Empty;
            foreach (NotificationTargetElement target in Configuration.NotificationTargets)
            {
                string pingUrl = target.TargetUrl + sitemapUrl;

                try
                {
                    HttpWebRequest.Create(pingUrl).GetResponse();
                    result += string.Format("Successfully pinged {0}. ",
                        target.Name);
                }
                catch (WebException ex)
                {
                    result += string.Format(
                        "Failed to ping {0}, error message: \"{1}\". ",
                        target.Name, ex.Message);
                }
            }

            return result.Trim();
        }

        private static SitemapUpdateNotificationConfiguration Configuration
        {
            get
            {
                return SitemapUpdateNotificationConfiguration.GetConfiguration();
            }
        }

        private static XNamespace SitemapNamespace
        {
            get
            {
                return SITEMAP_NAMESPACE;
            }
        }
    }
}
